#include "dictionary.h"
#include "model.h"

#include <fstream>
#include <iostream>
#include <random>
#include <set>
#include <sstream>
#include <stack>

#include <math.h>

#include <boost/algorithm/string/predicate.hpp>
#include <boost/date_time/local_time/local_time.hpp>
#include "boost/date_time/posix_time/posix_time.hpp"
#include "boost/date_time/posix_time/posix_time_types.hpp"
#include <boost/lexical_cast.hpp>


static unsigned int MAX_LOOP_COUNTER = 50;
static int MAX_LITERAL_WORDS = 25;

void predicate_m_t::init(string label, LITERAL_TYPES::enum_t literal_type){
    _label = label;
    _literal_type = literal_type;
    switch (_literal_type){
        case LITERAL_TYPES::INTEGER:{
            _range_min = boost::lexical_cast<string>(numeric_limits<unsigned short>::min());
            _range_max = boost::lexical_cast<string>(numeric_limits<unsigned short>::max());
            break;
        }
        case LITERAL_TYPES::STRING:{
            _range_min = string("A");
            _range_max = string("z");
            break;
        }
        case LITERAL_TYPES::DATE:{
            boost::posix_time::ptime cur_time (boost::posix_time::second_clock::local_time());
            boost::gregorian::date cur_date = cur_time.date();
            _range_min = string("1970-01-01");
            _range_max = boost::gregorian::to_iso_extended_string(cur_date);
            break;
        }
    }
    _distribution_type = DISTRIBUTION_TYPES::UNIFORM;
}

predicate_m_t::predicate_m_t (string label, LITERAL_TYPES::enum_t literal_type){
    init(label, literal_type);
}

predicate_m_t::predicate_m_t (string label, LITERAL_TYPES::enum_t literal_type, string range_min, string range_max){
    init(label, literal_type);
    _range_min = range_min;
    _range_max = range_max;
}

predicate_m_t::predicate_m_t (string label, LITERAL_TYPES::enum_t literal_type, string range_min, string range_max, DISTRIBUTION_TYPES::enum_t distribution_type){
    init(label, literal_type);
    _range_min = range_min;
    _range_max = range_max;
    _distribution_type = distribution_type;
}

predicate_m_t::predicate_m_t (const predicate_m_t & rhs){
    _label = rhs._label;
    _literal_type = rhs._literal_type;
    _range_min = rhs._range_min;
    _range_max = rhs._range_max;
    _distribution_type = rhs._distribution_type;
}

predicate_m_t * predicate_m_t::parse (const string & line){
    string label;
    LITERAL_TYPES::enum_t literal_type = LITERAL_TYPES::UNDEFINED;
    string range_min;
    string range_max;
    DISTRIBUTION_TYPES::enum_t distribution_type = DISTRIBUTION_TYPES::UNDEFINED;

    stringstream parser(line);
    int index = 0;
    string token;

    while (parser>>token){
        switch (index){
            case 0:{
                if (token.compare("#predicate")!=0){
                    cerr<<"[predicate_m_t::parse()]\tExpecting #predicate..."<<"\n";
                    exit(0);
                }
                break;
            }
            case 1:{
                label = token;
                break;
            }
            case 2:{
                if (token.compare("integer")==0 || token.compare("INTEGER")==0){
                    literal_type = LITERAL_TYPES::INTEGER;
                } else if (token.compare("string")==0 || token.compare("STRING")==0){
                    literal_type = LITERAL_TYPES::STRING;
                } else if (token.compare("date")==0 || token.compare("DATE")==0){
                    literal_type = LITERAL_TYPES::DATE;
                }
                break;
            }
            case 3:{
                range_min = token;
                break;
            }
            case 4:{
                range_max = token;
                break;
            }
            case 5:{
                if (token.compare("uniform")==0 || token.compare("UNIFORM")==0){
                    distribution_type = DISTRIBUTION_TYPES::UNIFORM;
                } else if (token.compare("normal")==0 || token.compare("NORMAL")==0){
                    distribution_type = DISTRIBUTION_TYPES::NORMAL;
                } else if (token.compare("zipfian")==0 || token.compare("ZIPFIAN")==0){
                    distribution_type = DISTRIBUTION_TYPES::ZIPFIAN;
                }
                break;
            }
        }
        index++;
    }

    if (index==3){
        return new predicate_m_t(label, literal_type);
    } else if (index==5){
        return new predicate_m_t(label, literal_type, range_min, range_max);
    } else if (index==6){
        return new predicate_m_t(label, literal_type, range_min, range_max, distribution_type);
    } else {
        cerr<<"[predicate_m_t::parse()]\tExpecting 2, 3 or 5 arguments..."<<"\n";
        exit(0);
    }
}

string predicate_m_t::generate (){
    string result = "";
    string literal = "";

    switch (_literal_type){
        case LITERAL_TYPES::INTEGER:{
            int min_value = boost::lexical_cast<int>(_range_min);
            int max_value = boost::lexical_cast<int>(_range_max);
            int interval = max_value - min_value;
            double r_value = model::generate_random(_distribution_type);
            int offset = ceil(r_value * interval);
            offset = (offset<0) ? 0 : offset;
            offset = (offset>interval) ? interval : offset;
            literal.append(boost::lexical_cast<string>(min_value + offset));
            break;
        }
        case LITERAL_TYPES::STRING:{
            pair<unsigned int, unsigned int> range = dictionary::get_instance()->get_interval(_range_min, _range_max);
            int interval = range.second - range.first - 1;
            double r_value = model::generate_random(_distribution_type);
            int offset = ceil(r_value * interval);
            offset = (offset<0) ? 0 : offset;
            offset = (offset>interval) ? interval : offset;
            literal.append(*(dictionary::get_instance()->get_word(range.first + offset)));
            // Keep appending a few more words from the dictionary...
            unsigned int wc = rand() % MAX_LITERAL_WORDS;
            unsigned int dict_size = dictionary::get_instance()->word_count();
            for (unsigned int index=0; index<wc; index++){
                literal.append(" ");
                literal.append(*(dictionary::get_instance()->get_word(rand()%dict_size)));
            }
            break;
        }
        case LITERAL_TYPES::DATE:{
            boost::posix_time::ptime min_time, max_time, gen_time;
            locale format (locale::classic(),new boost::posix_time::time_input_facet("%Y-%m-%d"));
            istringstream min_iss (_range_min);
            min_iss.imbue(format);
            min_iss>>min_time;
            istringstream max_iss (_range_max);
            max_iss.imbue(format);
            max_iss>>max_time;
            boost::posix_time::time_duration range (max_time - min_time);
            long interval = range.total_seconds();
            double r_value = model::generate_random(_distribution_type);
            long offset = ceil(r_value * interval);
            offset = (offset<0) ? 0 : offset;
            offset = (offset>interval) ? interval : offset;
            gen_time = min_time + boost::gregorian::days(offset/(24*3600));
            boost::gregorian::date gen_date = gen_time.date();
            literal.append(boost::gregorian::to_iso_extended_string(gen_date));
            break;
        }
    }

    result.append("<");
    result.append(_label);
    result.append(">");
    result.append("\t");
    result.append("\"");
    result.append(literal);
    result.append("\"");
    return result;
}

predicate_group_m_t::predicate_group_m_t (){
    _gen_probability = 1.0;
}

predicate_group_m_t::predicate_group_m_t (float gen_probability){
    _gen_probability = gen_probability;
}

predicate_group_m_t::predicate_group_m_t (const predicate_group_m_t & rhs){
    _gen_probability = rhs._gen_probability;
    for (unsigned int i=0; i<rhs._predicate_array.size(); i++){
        _predicate_array.push_back(new predicate_m_t(*(rhs._predicate_array[i])));
    }
}

predicate_group_m_t::~predicate_group_m_t(){
    for (unsigned int i=0; i<_predicate_array.size(); i++){
        delete _predicate_array[i];
    }
}

predicate_group_m_t * predicate_group_m_t::parse (const string & line){
    float gen_probability = 1.0;

    stringstream parser(line);
    int index = 0;
    string token;
    while (parser>>token){
        switch (index){
            case 0:{
                if (token.compare("#pgroup")!=0){
                    cerr<<"[predicate_group_m_t::parse()]\tExpecting #pgroup..."<<"\n";
                    exit(0);
                }
                break;
            }
            case 1:{
                gen_probability = boost::lexical_cast<float>(token);
                break;
            }
        }
        index++;
    }

    if (index==1){
        return new predicate_group_m_t();
    } else if (index==2){
        return new predicate_group_m_t(gen_probability);
    } else {
        cerr<<"[predicate_group_m_t::parse()]\tExpecting 0 or 1 arguments..."<<"\n";
        exit(0);
    }
}

resource_m_t::resource_m_t (string type_prefix, unsigned int scaling_coefficient){
    _type_prefix = type_prefix;
    _scaling_coefficient = scaling_coefficient;
}

resource_m_t::resource_m_t (const resource_m_t & rhs){
    _type_prefix = rhs._type_prefix;
    _scaling_coefficient = rhs._scaling_coefficient;
    for (unsigned int i=0; i<rhs._predicate_group_array.size(); i++){
        _predicate_group_array.push_back(new predicate_group_m_t(*(rhs._predicate_group_array[i])));
    }
}

resource_m_t::~resource_m_t (){
    for (unsigned int i=0; i<_predicate_group_array.size(); i++){
        delete _predicate_group_array[i];
    }
}

void resource_m_t::generate (map<string, unsigned int> & id_cursor_map, vector<string> & triple_lines){
    if (id_cursor_map.find(_type_prefix)==id_cursor_map.end()){
        id_cursor_map[_type_prefix] = 0;
    }
    for (unsigned int id=id_cursor_map[_type_prefix]; id<(id_cursor_map[_type_prefix] + _scaling_coefficient); id++){
        string subject = "";
        subject.append("<");
        subject.append(_type_prefix);
        subject.append(boost::lexical_cast<string>(id));
        subject.append(">");

        for (vector<predicate_group_m_t*>::const_iterator itr2=_predicate_group_array.begin(); itr2!=_predicate_group_array.end(); itr2++){
            predicate_group_m_t * predicate_group = *itr2;
            float draw = ((float) rand())/((float)RAND_MAX);
            if (draw<=predicate_group->_gen_probability){
                for (vector<predicate_m_t*>::const_iterator itr3=predicate_group->_predicate_array.begin(); itr3!=predicate_group->_predicate_array.end(); itr3++){
                    predicate_m_t * predicate = *itr3;
                    string triple = "";
                    triple.append(subject);
                    triple.append("\t");
                    triple.append(predicate->generate());
                    triple_lines.push_back(triple);
                }
            }
        }
    }
    id_cursor_map[_type_prefix] += _scaling_coefficient;
}

resource_m_t * resource_m_t::parse (const string & line){
    string type_prefix ("not_defined");
    unsigned int scaling_coefficient = 0;

    stringstream parser(line);
    int index = 0;
    string token;
    while (parser>>token){
        switch (index){
            case 0:{
                if (token.compare("#type")!=0){
                    cerr<<"[resource_m_t::parse()]\tExpecting #type..."<<"\n";
                    exit(0);
                }
                break;
            }
            case 1:{
                type_prefix = token;
                break;
            }
            case 2:{
                scaling_coefficient = boost::lexical_cast<unsigned int>(token);
                break;
            }
        }
        index++;
    }

    if (index==3){
        return new resource_m_t(type_prefix, scaling_coefficient);
    } else {
        cerr<<"[resource_m_t::parse()]\tExpecting 2 arguments..."<<"\n";
        exit(0);
    }

}

void association_m_t::init (string subject_type, string predicate, string object_type){
    _subject_type = subject_type;
    _predicate = predicate;
    _object_type = object_type;
    _left_cardinality = 1;
    _right_cardinality = 1;
    _left_cover = 1.0;
    _right_distribution = DISTRIBUTION_TYPES::UNIFORM;
}

association_m_t::association_m_t (string subject_type, string predicate, string object_type){
    init (subject_type, predicate, object_type);
}

association_m_t::association_m_t (string subject_type, string predicate, string object_type, unsigned int left_cardinality, unsigned int right_cardinality){
    init (subject_type, predicate, object_type);
    _left_cardinality = left_cardinality;
    _right_cardinality = right_cardinality;
}

association_m_t::association_m_t (string subject_type, string predicate, string object_type, unsigned int left_cardinality, unsigned int right_cardinality, float left_cover){
    init (subject_type, predicate, object_type);
    _left_cardinality = left_cardinality;
    _right_cardinality = right_cardinality;
    _left_cover = left_cover;
}

association_m_t::association_m_t (string subject_type, string predicate, string object_type, unsigned int left_cardinality, unsigned int right_cardinality, float left_cover, DISTRIBUTION_TYPES::enum_t right_distribution){
    init (subject_type, predicate, object_type);
    _left_cardinality = left_cardinality;
    _right_cardinality = right_cardinality;
    _left_cover = left_cover;
    _right_distribution = right_distribution;
}

void association_m_t::generate (map<string, unsigned int> & id_cursor_map, vector<string> & triple_lines){
    if (id_cursor_map.find(_subject_type)==id_cursor_map.end()){
        cerr<<"[association_m_t::parse()] Error: association cannot be defined over undefined resource '"<<_subject_type<<"'..."<<"\n";
        exit(0);
    }
    if (id_cursor_map.find(_object_type)==id_cursor_map.end()){
        cerr<<"[association_m_t::parse()] Error: association cannot be defined over undefined resource '"<<_object_type<<"'..."<<"\n";
        exit(0);
    }

    unsigned int left_instance_count = id_cursor_map[_subject_type];
    unsigned int right_instance_count = id_cursor_map[_object_type];
    set<unsigned int> mapped_instances;
    for (unsigned int left_id=0; left_id<left_instance_count; left_id++){
        float pr = ((float) rand()) / ((float) RAND_MAX);
        if (pr<=_left_cover){
            for (unsigned int j=0; j<_right_cardinality; j++){
                unsigned int loop_counter = 0;
                unsigned int right_id = 0;
                do {
                    double r_value = model::generate_random(_right_distribution);
                    right_id = ceil(r_value * right_instance_count);
                    right_id = (right_id>=right_instance_count) ? (right_instance_count-1) : right_id;
                    loop_counter++;
                } while (mapped_instances.find(right_id)!=mapped_instances.end() && loop_counter<MAX_LOOP_COUNTER);
                if (loop_counter<MAX_LOOP_COUNTER){
                    if (_left_cardinality==1){
                        mapped_instances.insert(right_id);
                    }
                    string subject(""), object(""), triple ("");
                    // FIXME:: You need to add replace-command...
                    subject.append("<");
                    subject.append(_subject_type);
                    subject.append(boost::lexical_cast<string>(left_id));
                    subject.append(">");

                    object.append("<");
                    object.append(_object_type);
                    object.append(boost::lexical_cast<string>(right_id));
                    object.append(">");

                    triple.append(subject);
                    triple.append("\t");
                    triple.append("<");
                    triple.append(_predicate);
                    triple.append(">");
                    triple.append("\t");
                    triple.append(object);
                    triple_lines.push_back(triple);
                } else {
                    cout<<"[association_m_t::generate] Warning:: failed to greedily satisfy cardinality constraints..."<<"\n";
                    cout<<"[association_m_t::generate] Ignoring association "
                        <<_subject_type<<left_id<<"-->"
                        <<_object_type<<right_id<<" after"<<MAX_LOOP_COUNTER<<" trials..."<<"\n";
                }
            }
        }
    }
}

association_m_t * association_m_t::parse (const map<string, unsigned int> & id_cursor_map, const string & line){
    string subject_type ("");
    string predicate ("");
    string object_type ("");
    unsigned left_cardinality = 1;
    unsigned right_cardinality = 1;
    float left_cover = 1.0;
    DISTRIBUTION_TYPES::enum_t right_distribution = DISTRIBUTION_TYPES::UNIFORM;

    stringstream parser(line);
    int index = 0;
    string token;
    while (parser>>token){
        switch (index){
            case 0: {
                if (token.compare("#association")!=0){
                    cerr<<"[association_m_t::parse()]\tExpecting #association..."<<"\n";
                    exit(0);
                }
                break;
            }
            case 1: {
                subject_type = token;
                break;
            }
            case 2: {
                predicate = token;
                break;
            }
            case 3: {
                object_type = token;
                break;
            }
            case 4: {
                left_cardinality = boost::lexical_cast<unsigned int>(token);
                break;
            }
            case 5: {
                right_cardinality = boost::lexical_cast<unsigned int>(token);
                break;
            }
            case 6: {
                left_cover = boost::lexical_cast<float>(token);
                break;
            }
            case 7: {
                if (token.compare("uniform")==0 || token.compare("UNIFORM")==0){
                    right_distribution = DISTRIBUTION_TYPES::UNIFORM;
                } else if (token.compare("normal")==0 || token.compare("NORMAL")==0){
                    right_distribution = DISTRIBUTION_TYPES::NORMAL;
                } else if (token.compare("zipfian")==0 || token.compare("ZIPFIAN")==0){
                    right_distribution = DISTRIBUTION_TYPES::ZIPFIAN;
                }
                break;
            }
        }
        index++;
    }
    if (index==4){
        return new association_m_t (subject_type, predicate, object_type);
    } else if (index==5){
        return new association_m_t (subject_type, predicate, object_type, left_cardinality, right_cardinality);
    } else if (index==6){
        return new association_m_t (subject_type, predicate, object_type, left_cardinality, right_cardinality, left_cover);
    } else if (index==7){
        return new association_m_t (subject_type, predicate, object_type, left_cardinality, right_cardinality, left_cover, right_distribution);
    }
}

model::model(const char * filename){
    srand (time(NULL));
    parse(filename);
}

model::~model(){
}

void model::generate (int scale_factor){
    vector<string> triples;
    for (int i=0; i<scale_factor; i++){
        for (vector<resource_m_t*>::iterator itr2=_resource_array.begin(); itr2!=_resource_array.end(); itr2++){
            resource_m_t * resource = *itr2;
            resource->generate(_id_cursor_map, triples);
        }
    }
    for (vector<association_m_t*>::iterator itr1=_association_array.begin(); itr1!=_association_array.end(); itr1++){
        association_m_t * association = *itr1;
        association->generate(_id_cursor_map, triples);
    }
    for (vector<string>::iterator itr1=triples.begin(); itr1!=triples.end(); itr1++){
        cout<<*itr1<<"\n";
    }
}

void model::parse (const char * filename){
    stack<pair<short,void*> > object_stack;
    ifstream fis (filename);
    string line;
    while (fis.good() && !fis.eof()){
        getline(fis, line);
        if (boost::starts_with(line, "#predicate")){
            predicate_m_t * predicate = predicate_m_t::parse(line);
            object_stack.push(pair<short,void*>(2, (void*)predicate));
        } else if (boost::starts_with(line, "#pgroup")){
            predicate_group_m_t * predicate_group = predicate_group_m_t::parse(line);
            object_stack.push(pair<short,void*>(1, (void*)predicate_group));
        } else if (boost::starts_with(line, "#/pgroup")){
            vector<predicate_m_t*> array;
            while (object_stack.top().first!=1){
                array.push_back((predicate_m_t * ) object_stack.top().second);
                object_stack.pop();
            }
            predicate_group_m_t * predicate_group = (predicate_group_m_t *) object_stack.top().second;
            predicate_group->_predicate_array.insert(predicate_group->_predicate_array.begin(), array.begin(), array.end());
        } else if (boost::starts_with(line, "#type")){
            resource_m_t * resource = resource_m_t::parse(line);
            object_stack.push(pair<short,void*>(0, (void*)resource));
        } else if (boost::starts_with(line, "#/type")){
            vector<predicate_group_m_t*> array;
            while (object_stack.top().first!=0){
                array.push_back((predicate_group_m_t * ) object_stack.top().second);
                object_stack.pop();
            }
            resource_m_t * resource = (resource_m_t *) object_stack.top().second;
            resource->_predicate_group_array.insert(resource->_predicate_group_array.begin(), array.begin(), array.end());
        } else if (boost::starts_with(line, "#association")){
            association_m_t * association = association_m_t::parse(_id_cursor_map, line);
            _association_array.push_back(association);
        }
    }
    fis.close();
    while (!object_stack.empty()){
        _resource_array.push_back((resource_m_t *) object_stack.top().second);
        object_stack.pop();
    }
}

double model::generate_random (DISTRIBUTION_TYPES::enum_t distribution_type){
    double result = 0.0;
    switch (distribution_type){
        case DISTRIBUTION_TYPES::UNIFORM:{
            result = ((double) rand()) / ((double) RAND_MAX);
            break;
        }
        case DISTRIBUTION_TYPES::NORMAL:{
            break;
        }
        case DISTRIBUTION_TYPES::ZIPFIAN:{
            break;
        }
        case DISTRIBUTION_TYPES::UNDEFINED:
        default:{
            break;
        }
    }
    result = (result<0.0) ? 0.0:result;
    result = (result>1.0) ? 1.0:result;
    return result;
}

int main(int argc, const char* argv[]) {
    if (argc==3){
        dictionary * dict = dictionary::get_instance();
        dict->init("/usr/share/dict/words");

        const char * filename = argv[1];
        unsigned int scale_factor = boost::lexical_cast<unsigned int>(string(argv[2]));
        model cur_model (filename);
        cur_model.generate(scale_factor);

        dictionary::destroy_instance();
    } else {
        cout<<"Usage:::\t./wsdobm <model-file> <scale-factor>"<<"\n";
    }
}
